#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>


void build_lbx(void);
void extract_lbx(void);
long read_offset(FILE * file);
long compute_offset(long cnt);
void fwrite_offset(FILE * file, long cnt);
void fwrite_file(FILE * file, long cnt, long bytes);
long extract_bits(long val, long b1, long b2);

//------------------------------------------------------------------------------------------------

int main(void)
{
    printf("LBXtool for 'Master of Orion' Version 2013.\n");
    printf("\n");
	char choice[100];

	choice[0] = '\0';
	while (choice[0] != 'B' && choice[0] != 'E' && choice[0] != 'X')
	{
	  printf("B=Build LBX Archive / E=Extract LBX Archive / X=Exit - make your choice:");
	  printf("\n");
	  scanf("%s", choice);
	  choice[0] = toupper(choice[0]);
	}
	if (choice[0] == 'B')
		build_lbx();
	else if (choice[0] == 'E')
		extract_lbx();
	else ;

	//printf("Exiting Tool.\n");
	return 0;
}

//------------------------------------------------------------------------------------------------

void extract_lbx(void)
{
	FILE * file;
	char filename[100];
	char outfilename[100];
	long n_files;
	FILE * outfile;
	long d1, d2;
	long byte_count = 0;
	long offsets[1000];
	long cnt;

	printf("Enter Filename of the lbx archive: ");
	scanf("%s", filename);
	file = fopen(filename, "rb");
	if (file == NULL)
	{
		printf("Error opening file.\n");
		return;
	}

	d1 = fgetc(file);
	if (feof(file)) return;
	d2 = fgetc(file);
	if (feof(file)) return;
	n_files = d2*256+d1;
	byte_count = 2;

	printf("Number of files to extract: %d\n", n_files);

	/* Read past LBX File Identifier & version # */
	fgetc(file);
	fgetc(file);
	fgetc(file);
	fgetc(file);  // End of LBX identifier = AD FE 00 00
	fgetc(file);
	fgetc(file);  // End of file version 00 00
	byte_count += 6;

	/* One extra offset for EOF */
	for (cnt = 0; cnt <= n_files; cnt++)
	{
	    //if (read_offset(file)==1) offsets[cnt] =temp_offset;
        offsets[cnt] = read_offset(file);
		if (feof(file))
		{
			printf("Error: Unexpected End of File\n");
			return;
		}
		byte_count += 4;

		printf("Offset: File #%d = %d\n", cnt, offsets[cnt]);
	}

	/* Read to first offset */
	while (byte_count < offsets[0])
	{
		fgetc(file);
		byte_count++;
		if (feof(file))
		{
			printf("Error: Unexpected End of File\n");
			return;
		}
	}

	/* Extract the files */
	cnt = 0;
	while (cnt < n_files)
	{
		sprintf(outfilename, "file%d.txt", cnt+1);
		outfile = fopen(outfilename, "wb");
		if (outfile == NULL)
		{
			printf("Error opening output file %s\n", outfilename);
			return;
		}
		while (byte_count < offsets[cnt+1])
		{
			fprintf(outfile, "%c", fgetc(file));
			byte_count++;
			if (feof(file))
			{
				printf("Error: Unexpected End of File\n");
				return;
			}
		}
		fclose(outfile);
		cnt++;
	}

	fclose(file);
	//scanf("%s",NULL);
	return;
}

//------------------------------------------------------------------------------------------------

long read_offset(FILE * file)
{
	long n1, n2, n3, n4;

	n1 = fgetc(file);
	if (feof(file)) return 0;
	n2 = fgetc(file);
	if (feof(file)) return 0;
	n3 = fgetc(file);
	if (feof(file)) return 0;
	n4 = fgetc(file);
	if (feof(file)) return 0;

    //printf("Offset: File = %d,%d,%d,%d\n",n4,n3,n2,n1);
    n1=(16777216*n4 + 65536*n3 + 256*n2 + n1);
	return n1;
}

//------------------------------------------------------------------------------------------------

void build_lbx(void)
{
	char filename[100];
	FILE * file;
	long n_files;
	long offsets[1000];
	long cnt;
	long d1, d2;
	long n_bytes;

	printf("Enter LBX archive name to build: \n");
	scanf("%s", filename);
	printf("# of files to include: ");
	scanf("%d", &n_files);
	if (n_files < 1 || n_files > 1000)
	{
		printf("# of files must be between 1 and 1000\n");
		return;
	}
	file = fopen(filename, "wb");
	if (file == NULL)
	{
		printf("Error opening %s for output.\n", filename);
		return;
	}

	/* Compute our offsets */
	/* Start of 1st file */
	/*  This wasn't working trying to always use 2048 like some of the files ive seen
	offsets[0] = 8 + (n_files+1)*4; // 8 bytes header + 4 bytes per offset
	*/
	//offsets[0] = 2048;
	offsets[0] = 1216;

	cnt = 1;
	while (cnt <= n_files)
	{
		offsets[cnt] = offsets[cnt-1] + compute_offset(cnt); // Start of previous offset
		cnt++;
	}

	/* Write header */
	d1 = n_files & 255;
	d2 = n_files & 65280;
	d2 = d2 >> 8;

	//fprintf(file, "%c%c%c%c%c%c%c%c", d1, d2, 173, 254, 0, 0, 0, 0);
	fprintf(file, "%c%c%c%c%c%c%c%c", d1, d2, 173, 254, 0, 0, 1, 0);

	/* Write out our n+1 offsets */
	cnt = 0;
	while (cnt <= n_files)
	{
		fwrite_offset(file, offsets[cnt]);
		cnt++;
	}

	/* write filler up to first offset */
	n_bytes = 8 + (n_files+1)*4;
	while ((unsigned) n_bytes < offsets[0])
	{
		fprintf(file, "%c", 0);
		n_bytes++;
	}

	/* now write out each of the files */
	cnt = 0;
	while (cnt < n_files)
	{
		n_bytes = offsets[cnt+1] - offsets[cnt];
		fwrite_file(file,  cnt, n_bytes);
		cnt++;
	}
	fclose(file);
	//scanf("%s",NULL);
	return;
}

//------------------------------------------------------------------------------------------------

void fwrite_file(FILE * file, long cnt, long bytes)
{
	FILE * outfile;
	char filename[100];
	long cnt2;

	sprintf(filename, "file%d.txt", (cnt+1));
	outfile = fopen(filename, "rb");
	if (outfile == NULL)
	{
		printf("Error opening file: %s\n", filename);
		exit (-1);
	}

	printf("Archiving: %s  (%d bytes)\n", filename, bytes);
	cnt2 = 0;
	while (cnt2 < bytes)
	{
		fprintf(file, "%c", fgetc(outfile));
		if (feof(outfile))
		{
			printf("Error:  Unexpected end of file.\n");
			exit (-1);
		}
		cnt2++;
	}
	fclose(outfile);
	return;
}

//------------------------------------------------------------------------------------------------

void fwrite_offset(FILE * file, long offset)
{
	long n1, n2, n3, n4;

    //4294967296  16777216  65536  256  1

	n1 = offset & 255;
	n2 = offset & 65280;
	n3 = offset & 16711680;
    n4 = offset & 4278255616;

	n2 = n2 >> 8;
	n3 = n3 >> 16;
	n4 = n4 >> 24;

	fprintf(file, "%c%c%c%c", n1, n2, n3, n4);
	return;
}

//------------------------------------------------------------------------------------------------

long compute_offset(long cnt)
{
	char filename[100];
	long byte_count = 0;
	FILE * file;

	sprintf(filename, "file%d.txt", cnt);
	file = fopen(filename, "r");
	if (file == NULL)
	{
		printf("Error opening file %s\n", filename);
		exit(-1);
	}

	/*
	while (!feof(file))
	{
		fgetc(file);
		if (!feof(file))
			byte_count++;
	}*/

    fseek(file, 0, SEEK_END);
    byte_count = ftell(file);
    rewind (file);

	printf("Size of file #%d = %d\n", cnt, byte_count);

	fclose(file);
	return byte_count;
}

